%{
#include <regex.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
int yylex();
int yyerror();
int yyparse();

typedef struct {
    char *s;
    int   n;
} Tok;
#define YYSTYPE Tok

char *tokstr(Tok tok, char *buf)
{
    char *p = tok.s;
    if (!p) sprintf(p = buf, "%d", tok.n);
    return p;
}

Tok match(Tok tstr, Tok tregx)
{
    char b1[21], b2[21];
    char *str  = tokstr(tstr , b1);
    char *regx = tokstr(tregx, b2);

    regex_t    re;
    regmatch_t matches[2];
    char       anchreg[strlen(regx) + 2];

    sprintf(anchreg, "^%s", regx);
    if (regcomp(&re, anchreg, 0))
        yyerror("regcomp failed");

    if (regexec(&re, str, 2, matches, 0))
        return (Tok){ NULL, 0 };

    if (re.re_nsub) {
        if (matches[1].rm_so == -1)
            return (Tok){ "", 0 };

        int   len = matches[1].rm_eo - matches[1].rm_so + 1;
        char *ret = malloc(len); // find a way to get rid of this...
        if (!ret)
            yyerror("malloc failed");
        strncpy(ret, str + matches[1].rm_so, len);
        return (Tok){ ret, 0 };
    }

    return (Tok){ NULL, matches[0].rm_eo - matches[0].rm_so };
}

int tokcmp(Tok a, Tok b)
{
    if (!a.s && !b.s)
        return (a.n > b.n) - (a.n < b.n);

    char b1[21], b2[21];
    char *p = tokstr(a, b1);
    char *q = tokstr(b, b2);

    // get LC_COLLATE and use strcoll?
    return strcmp(p, q);
}

void num(Tok t)
{
    if (t.s)
        yyerror(t.s);
}

%}
%token TOK EOA
%token GE LE NE

%left '|'
%left '&'
%left '=' '>' GE '<' LE NE
%left '+' '-'
%left '*' '/' '%'
%left ':'
%%

list: expr EOA { if ($1.s) printf("%s\n", $1.s);
                 else      printf("%d\n", $1.n);
                 return !(($1.s && *$1.s) || $1.n); }
    ;

expr: TOK
    | expr '|' expr { if      ( $1.s && *$1.s) $$ = (Tok){ $1.s,    0 };
                      else if (!$1.s &&  $1.n) $$ = (Tok){ NULL, $1.n };
                      else if ( $3.s && *$3.s) $$ = (Tok){ $3.s,    0 };
                      else                     $$ = (Tok){ NULL, $3.n }; }

    | expr '&' expr { if ((($1.s && *$1.s) || $1.n) &&
                          (($3.s && *$3.s) || $3.n)) $$ = $1;
                      else $$ = (Tok){ NULL, 0 }; }

    | expr '=' expr { $$ = (Tok){ NULL, tokcmp($1, $3) == 0 }; }
    | expr '>' expr { $$ = (Tok){ NULL, tokcmp($1, $3) >  0 }; }
    | expr GE  expr { $$ = (Tok){ NULL, tokcmp($1, $3) >= 0 }; }
    | expr '<' expr { $$ = (Tok){ NULL, tokcmp($1, $3) <  0 }; }
    | expr LE  expr { $$ = (Tok){ NULL, tokcmp($1, $3) <= 0 }; }
    | expr NE  expr { $$ = (Tok){ NULL, tokcmp($1, $3) != 0 }; }

    | expr '+' expr { num($1); num($3); $$ = (Tok){ NULL, $1.n + $3.n }; }
    | expr '-' expr { num($1); num($3); $$ = (Tok){ NULL, $1.n - $3.n }; }
    | expr '*' expr { num($1); num($3); $$ = (Tok){ NULL, $1.n * $3.n }; }
    | expr '/' expr { num($1); num($3); $$ = (Tok){ NULL, $1.n / $3.n }; }
    | expr '%' expr { num($1); num($3); $$ = (Tok){ NULL, $1.n % $3.n }; }

    | expr ':' expr { $$ = match($1, $3); }

    | '(' expr ')'  { $$ = $2; }
    ;

%%
char **args;

int main(int argc, char **argv)
{
    args = argv + 1;
    return yyparse();
}

int yylex(void)
{
    char *p = *args++, *q;
    if (!p)
        return EOA;

    int d = strtol(p, &q, 0);
    if (*p && !*q) {
        yylval = (Tok){ NULL, d };
        return TOK;
    }

    char ops[] = "|&=><+-*/%():";
    if (p[0] && !p[1] && strchr(ops, *p))
        return *p;

    if (p[0] && p[1] == '=')
        switch (*p) {
            case '>': return GE;
            case '<': return LE;
            case '!': return NE;
        }

    yylval = (Tok){ p, 0 };
    return TOK;
}

int yyerror(char *s)
{
    printf("error: %s\n", s);
    exit(2);
}
